/*
 * CS 106X Fall 2018 - Nick Troccoli
 * C++ solutions for problems from Section #6 (Week 7)
 * Section topic: Binary Trees
 */

#include <iostream>
#include <fstream>
#include "console.h"
#include "strlib.h"
#include "section6-solutions.h"

using namespace std;

/**********************************
 * 1. Traversals - (binary trees) *
 **********************************/

// no code required; see section solution handout


/*************************************
 * 2. BST insertion - (binary trees) *
 *************************************/

// no code required; see section solution handout


/*****************************************
 * 3. countLeftNodes - (binary trees) *
 *****************************************/

int countLeftNodes(TreeNode* node) {
    if (node == nullptr) {
        return 0;
    } else if (node->left == nullptr) {
        return countLeftNodes(node->right);
    } else {
        return 1 + countLeftNodes(node->left) + countLeftNodes(node->right);
    }
}

/*****************************
 * 4. isBST - (binary trees) *
 *****************************/
// approach 1 -- set bounds on valid data values and restrict them while walking down tree
bool isBSTHelper1(TreeNode* node, int min, int max) {
    if(node == nullptr) return true;
    if(node->data < min || node->data > max) return false;
    return isBSTHelper1(node->left, min, node->data - 1) &&
           isBSTHelper1(node->right, node->data + 1, max);
}

bool isBST1(TreeNode* root) {
    return isBSTHelper1(root, INT_MIN, INT_MAX);
}

// approach 2 -- use prev pointers to walk back up and across the tree
bool isBSTHelper2(TreeNode* node, TreeNode*& prev) {   // an in-order walk
    if (node == nullptr) {                             // of the tree, storing the
        return true;                                   // last visited node in 'prev'
    } else if (!isBSTHelper2(node->left, prev) || (prev && node->data <= prev->data)) {
        return false;
    } else {
        prev = node;
        return isBSTHelper2(node->right, prev);
    }
}

bool isBST2(TreeNode* node) {
    TreeNode* prev = nullptr;
    return isBSTHelper2(node, prev);
}


/************************************
 * 5. removeLeaves - (binary trees) *
 ************************************/

void removeLeaves(TreeNode*& node) {
    if (node != nullptr) {
        if (node->left == nullptr && node->right == nullptr) {
            delete node;
            node = nullptr;
        } else {
            removeLeaves(node->left);
            removeLeaves(node->right);
        }
    }
}


/****************************************
 * 6. completeToLevel - (linked lists) *
 ****************************************/

void completeToLevelHelper(TreeNodeString*& node, int k, int currLevel) {
    if (currLevel <= k) {
        if (node == nullptr) {
            node = new TreeNodeString("??");
        }
        completeToLevelHelper(node->left,  k, currLevel + 1);
        completeToLevelHelper(node->right, k, currLevel + 1);
    }
}

void completeToLevel(TreeNodeString*& node, int k) {
    if (k < 1) {
        throw k;
    } else {
        completeToLevelHelper(node, k, 1);
    }
}


/**********************************
 * 7. wordExists - (binary trees) *
 **********************************/

// helper to search the given subtree for the rest of the string
bool suffixExists(TreeNodeChar* node, const string& suffix) {
    if (suffix.empty()) {
        return true;
    } else if (node == nullptr) {
        return false;
    } else if (suffix[0] != node->data) {
        return false;
    } else {
        return suffixExists(node->left, suffix.substr(1))
                || suffixExists(node->right, suffix.substr(1));
    }
}

bool wordExists(TreeNodeChar* node, const string& str) {
    if (str.empty()) {
        return true;   // the empty tree contains the empty string
    } else if (node == nullptr) {
        return false;
    }  else if (suffixExists(node, str)) {
        return true;
    } else {
        return wordExists(node->left, str) || wordExists(node->right, str);
    }
}


/* main function for testing individual problems */
/* note that the printSideways is defined in section6-solutions.h */
int main() {
    cout << "Problem 3 (countLeftNodes):" << endl;

    /* CHALLENGE TO THOSE WHO ARE READING THIS: find and write a way to construct a binary tree that
     *                                          is more convenient than the way it's done below, and
     *                                          then email it to Zach at zacharyb@stanford.edu. :)
     */

    TreeNode* root3 = new TreeNode(3,
                                   new TreeNode(5, //left
                                                new TreeNode(1, //left
                                                             nullptr, //left
                                                             nullptr  //right
                                                             ),
                                                nullptr //right
                                                ),
                                   new TreeNode(2, //right
                                                new TreeNode(4, //left
                                                             nullptr, //left
                                                             nullptr  //right
                                                             ),
                                                new TreeNode(6, //right
                                                             nullptr, //left
                                                             nullptr  //right
                                                             )
                                                )
                                   );
    printSideways<TreeNode>(root3);
    cout << "Left nodes of above tree: " << countLeftNodes(root3) << endl;

    cout << endl << "Problem 4 (isBST):" << endl;
    TreeNode* root4First = new TreeNode(7,
                                        new TreeNode(4,
                                                     nullptr,
                                                     nullptr
                                                     ),
                                        new TreeNode(9,
                                                     new TreeNode(8,
                                                                  nullptr,
                                                                  nullptr
                                                                  ),
                                                     new TreeNode(10,
                                                                  nullptr,
                                                                  nullptr
                                                                  )
                                                     )
                                        );
    printSideways<TreeNode>(root4First);
    cout << "Above tree is BST:" << endl;
    cout << "Approach 1: " << boolToString(isBST1(root4First)) << endl;
    cout << "Approach 2: " << boolToString(isBST2(root4First)) << endl << endl;

    TreeNode* root4Second = new TreeNode(4,
                                         new TreeNode(3,
                                                      nullptr,
                                                      nullptr
                                                      ),
                                         new TreeNode(9,
                                                      new TreeNode(6,
                                                                   nullptr,
                                                                   nullptr
                                                                   ),
                                                      new TreeNode(8,
                                                                   nullptr,
                                                                   nullptr
                                                                   )
                                                      )
                                         );
    printSideways<TreeNode>(root4Second);
    cout << "Above tree is BST:" << endl;
    cout << "Approach 1: " << boolToString(isBST1(root4Second)) << endl;
    cout << "Approach 2: " << boolToString(isBST2(root4Second)) << endl;


    cout << endl << "Problem 5 (removeLeaves):" << endl;
    TreeNode* root5 = new TreeNode(7,
                                   new TreeNode(3,
                                                new TreeNode(1,
                                                             nullptr,
                                                             nullptr),
                                                new TreeNode(4,
                                                             nullptr,
                                                             nullptr)
                                                ),
                                   new TreeNode(9,
                                                new TreeNode(6,
                                                             nullptr,
                                                             nullptr
                                                             ),
                                                new TreeNode(8,
                                                             nullptr,
                                                             new TreeNode(0,
                                                                          nullptr,
                                                                          nullptr)
                                                             )
                                                )
                                   );

    cout << "Before:" << endl;
    printSideways(root5);
    removeLeaves(root5);
    cout << "After first call:" << endl;
    printSideways(root5);
    removeLeaves(root5);
    cout << "After second call:" << endl;
    printSideways(root5);
    removeLeaves(root5);
    cout << "After third call:" << endl;
    printSideways(root5);
    removeLeaves(root5);
    cout << "After fourth call:" << endl;
    printSideways(root5);

    cout << endl << "Problem 6 (completeToLevel):" << endl;
    TreeNodeString* root6 = new TreeNodeString("sahami",
                                               nullptr,
                                               new TreeNodeString("stepp",
                                                                  nullptr,
                                                                  nullptr
                                                                  )
                                               );
    completeToLevel(root6, 3);
    printSideways<TreeNodeString>(root6);


    cout << "Problem 7 (wordExists):" << endl;
    TreeNodeChar* root7 = new TreeNodeChar('c',
                                           new TreeNodeChar('r',
                                                            new TreeNodeChar('a',
                                                                             new TreeNodeChar('b',
                                                                                              nullptr,
                                                                                              nullptr
                                                                                              ),
                                                                             new TreeNodeChar('n',
                                                                                              nullptr,
                                                                                              nullptr
                                                                                              )
                                                                             ),
                                                            new TreeNodeChar('u',
                                                                             new TreeNodeChar('g',
                                                                                              new TreeNodeChar('h',
                                                                                                               nullptr,
                                                                                                               nullptr),
                                                                                              nullptr
                                                                                              ),
                                                                             new TreeNodeChar('s',
                                                                                              nullptr,
                                                                                              new TreeNodeChar('t',
                                                                                                               nullptr,
                                                                                                               nullptr)
                                                                                              )
                                                                             )
                                                            ),
                                           new TreeNodeChar('z',
                                                            new TreeNodeChar('a',
                                                                             nullptr,
                                                                             new TreeNodeChar('r',
                                                                                              nullptr,
                                                                                              nullptr)
                                                                             ),
                                                            new TreeNodeChar('e',
                                                                             new TreeNodeChar('r',
                                                                                              nullptr,
                                                                                              new TreeNodeChar('o',
                                                                                                               nullptr,
                                                                                                               nullptr)
                                                                             ),
                                                                             new TreeNodeChar('d',
                                                                                              nullptr,
                                                                                              nullptr)
                                                                             )
                                                            )
                                           );

    printSideways(root7);
    cout << "\"crust\" exists in the above tree: " << boolToString(wordExists(root7, "crust")) << endl;
    cout << "\"rug\" exists in the above tree: " << boolToString(wordExists(root7, "rug")) << endl;
    cout << "\"arc\" exists in the above tree: " << boolToString(wordExists(root7, "arc")) << endl;
    cout << "\"cut\" exists in the above tree: " << boolToString(wordExists(root7, "cut")) << endl;


    return 0;
}
